@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-import pytest
+import pytest, re
 from copy import copy
 from collections.abc import Mapping, Iterable
 from contextlib2 import nullcontext
@@ -73,7 +73,7 @@ class Namespace(Mapping):
             >>> with_math.eval('sqrt(4)')
             2.0
 
-        Once you have an initialized a namespace, you can use it to...
+        Once you have an initialized namespace, you can use it to...
 
         ...evaluate expressions:
 
@@ -182,13 +182,19 @@ class Namespace(Mapping):
                 defer actually evaluating it until inside the test function.
 
         Returns:
-            Any: If any *src* expressions were given: The result(s) of evaluating 
-            the given expression(s).
-
-            If no *src* expressions were given: A `functools.partial` version 
-            of this function that will remember any specified keyword 
+            Any: If no *src* expressions were given: A `functools.partial` 
+            version of this function that will remember any specified keyword 
             arguments.
 
+            If any *src* expressions were given and the *defer* argument was 
+            True: A `functools.partial` version of this function that will 
+            remember the expressions and keyword arguments given to it.  The 
+            return value will also have an `eval` attribute that aliases the 
+            deferred function.
+
+            If any *src* expressions were given and the *defer* argument was 
+            False: The result(s) of evaluating the given expression(s).
+
         Examples:
             Basic usage::
 
@@ -208,24 +214,24 @@ class Namespace(Mapping):
                 
                 >>> @parametrize_from_file(  # doctest: +SKIP
                 ...         schema={
-                ...             'expr': with_a.eval(defer=True),
+                ...             'expected': with_a.eval(defer=True),
                 ...         },
                 ... )
-                ... def test_snippet(expr)
-                ...     expr = expr()
+                ... def test_snippet(expected)
+                ...     expected = expected.eval()
 
         Details:
             `unittest.mock.Mock` instances are handled specially by this 
             method.  Specifically, they are returned unchanged (i.e. without 
             being evaluated).  This special case exists because 
-            `voluptuous.Namespace.error_or` uses
-            `unittest.mock.Mock` instances as placeholders when an exception is 
-            expected.
+            `voluptuous.Namespace.error_or` uses `unittest.mock.Mock` instances 
+            as placeholders when an exception is expected.
         """
         if not src:
             return partial(self.eval, keys=keys, defer=defer)
         if defer:
-            return partial(self.eval, *src, keys=keys)
+            f = f.eval = partial(self.eval, *src, keys=keys)
+            return f
 
         src = src[0] if len(src) == 1 else list(src)
         recurse = partial(self.eval, keys=keys)
@@ -260,16 +266,28 @@ class Namespace(Mapping):
                 snippet.  Whatever that function returns will be passed on to 
                 the caller.
 
+            defer (bool):
+                If true, instead of actually evaluating the given snippet, 
+                return a no-argument callable that will evaluate it when 
+                called.  The purpose of this is to make it possible to specify 
+                that a test parameter should be executed in a schema, but to 
+                defer actually executing it until inside the test function.
+
         Returns:
-            Any: If *src* was specified and *get* was not: A new `Namespace` 
+            Any: If *src* was not specified: A `functools.partial` version of 
+            this function that will remember any specified keyword arguments.
+
+            If *src* was specified and the *defer* argument was True: A 
+            `functools.partial` version of this function that will remember the 
+            snippet and keyword arguments given to it.  The return value will 
+            also have an `exec` attribute that aliases the deferred function.
+
+            If *src* was specified and *get* was not: A new `Namespace` 
             containing all of the variables defined in the snippet.
 
             If *src* and *get* were specified: The value(s) of the indicated 
             variable(s) defined the snippet.
 
-            If *src* was not specified: A `functools.partial` version of this 
-            function that will remember any specified keyword arguments.
-
         Examples:
             Basic usage::
 
@@ -294,7 +312,7 @@ class Namespace(Mapping):
                 ...         },
                 ... )
                 ... def test_snippet(snippet)
-                ...     snippet = snippet()
+                ...     snippet = snippet.exec()
 
         Details:
             `unittest.mock.Mock` instances are handled specially by this 
@@ -307,7 +325,8 @@ class Namespace(Mapping):
         if src is SENTINEL:
             return partial(self.exec, get=get, defer=defer)
         if defer:
-            return partial(self.exec, src, get=get)
+            f = f.exec = partial(self.exec, src, get=get)
+            return f
         if isinstance(src, Mock):
             return src
 
@@ -369,6 +388,12 @@ class Namespace(Mapping):
                   available via the *exc* variable.  This field is typically 
                   used to make free-form assertions about the exception object.
 
+                - "cause" (optional): A string that should evaluate to an 
+                  integer 'n'.  All other checks will apply to the n-th direct 
+                  cause of the caught exception, instead of the exception 
+                  itself.  This is useful in cases where the exception you want 
+                  to test is wrapped in some other exception.
+
                 Note that everything is expected to be strings, because this 
                 method is meant to help with parsing exception information from 
                 a text file, e.g. in the NestedText_ format.  All evaluations 
@@ -417,7 +442,8 @@ class Namespace(Mapping):
             err.messages = require_list(params.get('message', []))
             err.patterns = require_list(params.get('pattern', []))
             err.attr_strs = params.get('attrs', {})
-            err.assertions_str = params.get('assertions', {})
+            err.assertions_str = params.get('assertions', '')
+            err.cause_str = params.get('cause', '')
 
         return err
 
@@ -438,13 +464,14 @@ class ExpectError:
     # The only way to support this is to implement the context manager from 
     # scratch.
 
-    def __init__(self, namespace, *, type_str=Exception, messages=[], patterns=[], attr_strs={}, assertions_str=''):
+    def __init__(self, namespace, *, type_str=Exception, messages=[], patterns=[], attr_strs={}, assertions_str='', cause_str=''):
         self.namespace = namespace
         self.type_str = type_str
         self.messages = messages
         self.patterns = patterns
         self.attr_strs = attr_strs
         self.assertions_str = assertions_str
+        self.cause_str = cause_str
 
     def __repr__(self):
         attrs = {
@@ -453,6 +480,7 @@ class ExpectError:
                 'patterns': self.patterns,
                 'attrs': self.attr_strs,
                 'assertions': self.assertions_str,
+                'cause': self.cause_str,
         }
         attr_str = ' '.join(
                 f'{k}={v!r}'
@@ -464,32 +492,42 @@ class ExpectError:
         return True
 
     def __enter__(self):
+        pass
+
+    def __exit__(self, exc_type, exc_value, exc_tb):
+        __tracebackhide__ = True
+
+        ns = self.namespace
         type = self.namespace.eval(self.type_str)
         if isinstance(type, list):
             type = tuple(type)
 
-        self.raises_cm = pytest.raises(type)
-        self.exc_info = self.raises_cm.__enter__()
+        assert exc_type is not None, f"DID NOT RAISE {type}"
 
-    def __exit__(self, *args):
-        if self.raises_cm.__exit__(*args):
-            ns = self.namespace
-            exc = self.exc_info.value
+        del exc_type, exc_tb
+        for i in range(ns.eval(self.cause_str or '0')):
+            assert exc_value.__cause__ is not None, f"{exc_value.__class__.__name__} has no direct cause"
+            exc_value = exc_value.__cause__
 
-            for msg in self.messages:
-                assert msg in str(exc), f'{msg!r} not in {str(exc)!r}'
+        if not isinstance(exc_value, type):
+            return False
 
-            for pat in self.patterns:
-                self.exc_info.match(pat)
+        exc_str = str(exc_value)
 
-            for attr, value_str in self.attr_strs.items():
-                assert hasattr(exc, attr)
-                assert getattr(exc, attr) == ns.eval(value_str)
+        for msg in self.messages:
+            assert msg in exc_str, f'{msg!r} not in {exc_str!r}'
 
-            if self.assertions_str:
-                ns.fork(err=exc).exec(self.assertions_str)
+        for pat in self.patterns:
+            assert re.search(pat, exc_str), "regex pattern {pat!r} does not match {exc_str!r}"
 
-            return True
+        for attr, value_str in self.attr_strs.items():
+            assert hasattr(exc_value, attr)
+            assert getattr(exc_value, attr) == ns.eval(value_str)
+
+        if self.assertions_str:
+            ns.fork(exc=exc_value).exec(self.assertions_str)
+
+        return True
 
 def star(module):
     """
