@@ -35,6 +35,7 @@ def _decorator_factory(api_func):
             path=None,
             key=None,
             *,
+            loaders=None,
             preprocess=None,
             schema=None,
             test_func=decopatch.DECORATED,
@@ -52,9 +53,11 @@ def _decorator_factory(api_func):
                 test_module=test_module,
                 test_path=test_path,
         ):
+            loaders = _override_global_loaders(loaders)
             param_names, param_values = load_parameters(
-                    _resolve_param_path(test_path, path),
+                    _resolve_param_path(test_path, path, loaders),
                     key or test_func.__name__,
+                    loaders=loaders,
                     preprocess=preprocess,
                     schema=schema,
             )
@@ -99,22 +102,33 @@ def parametrize(param_names, param_values, kwargs):
             of paths with the same length as this argument (meaning: look up 
             each key in the corresponding file).
 
+        loaders (dict):
+            A dictionary mapping file extensions (e.g. `'.nt'`) to functions 
+            that load test parameters from files (e.g. `nt.load`).  See 
+            `add_loader` for a more complete description of these functions.  
+            The items in this dictionary override the globally defined loaders 
+            for the purposes of the decorated test function.
+
         preprocess (collections.abc.Callable):
             A function that will be allowed to modify the list of test cases 
             loaded from the parameter file, e.g. to programmatically generate 
             and/or prune test cases.  The function should have the following 
             signature::
 
-                def preprocess(params: Any) -> List[Dict[str, Any]]
+                def preprocess(params: Any, [context: Context]) -> List[Dict[str, Any]]
 
-            The argument will be the value associated with the given key in the 
-            parameter file.  Typically this value is a list, but it could be 
-            anything.  If multiple parameter files and/or keys are specified, 
-            this function is called separately on each one.  The return value 
-            must be a list of dicts.  Each of these dicts will be further 
-            processed by the *schema* argument before being used to parametrize 
-            the test function.  Note that this function does get access to the 
-            special *id* and *marks* fields, unlike the *schema* function.
+            The first argument will be the value associated with the given key 
+            in the parameter file.  Typically this value is a list, but it 
+            could be anything.  If the function accepts a second argument 
+            (which it doesn't have to), it will be an object with *path* and 
+            *key* attributes specifying where the aforementioned value was 
+            loaded from.  If multiple parameter files and/or keys are 
+            specified, this function will be called separately on each one.  
+            The return value must be a list of dicts.  Each of these dicts will 
+            be further processed by the *schema* argument before being used to 
+            parametrize the test function.  Note that this function does get 
+            access to the special *id* and *marks* fields, unlike the *schema* 
+            function.
 
         schema (collections.abc.Callable):
             A function that will be used to validate and/or transform each set 
@@ -190,7 +204,7 @@ def parametrize(param_names, param_values, kwargs):
       them by appending a numeric id that counts up from 0.
     
     :param str,list marks:
-      One or more :doc:`marks <mark>` (like `skip <pytest.mark.skip>` or `xfail 
+      One or more :doc:`marks <how-to/mark>` (like `skip <pytest.mark.skip>` or `xfail 
       <pytest.mark.xfail>`) that should be applied this particular set of 
       parameters.  If a plain string is given, it will be split on commas.  If 
       a list is given, each element should be a string.  Pytest will issue a 
@@ -255,7 +269,7 @@ def fixture(param_names, param_values, kwargs):
         kwargs:
             See :deco:`parametrize`.
 
-    :doc:`Fixtures <fixture>` allow multiple test functions to be initialized 
+    :doc:`Fixtures <reference/fixtures>` allow multiple test functions to be initialized 
     in the same way.  Fixtures can be parametrized, in which case the fixture 
     will be reinitialized (and any dependent test functions rerun) for each 
     parameter.  
@@ -284,6 +298,7 @@ def load_parameters(
         path,
         key,
         *,
+        loaders=None,
         preprocess=None,
         schema=None,
     ):
@@ -300,6 +315,9 @@ def load_parameters(
         key (str, list):
             See: :deco:`parametrize`
 
+        loaders (dict):
+            See: :deco:`parametrize`
+
         preprocess (collections.abc.Callable):
             See: :deco:`parametrize`
 
@@ -319,6 +337,7 @@ def load_parameters(
     apply them all to the same test function.
     """
     test_params = []
+    loaders = _override_global_loaders(loaders)
 
     try:
         for path_i, key_i in zip_broadcast(path, key, strict=True):
@@ -326,13 +345,14 @@ def load_parameters(
                     "parameter file: {param_path}",
                     param_path=path_i,
             ):
-                p = _load_test_params(path_i, key_i)
+                p = _load_test_params(loaders, path_i, key_i)
+                context = Context(path_i, key_i)
 
                 with ConfigError.add_info(
                         "top-level key: {key}",
                         key=key_i,
                 ):
-                    p = _process_test_params(p, preprocess, schema)
+                    p = _process_test_params(p, preprocess, context, schema)
                     test_params += p
 
     except UnequalIterablesError:
@@ -348,7 +368,13 @@ def load_parameters(
     return _init_parametrize_args(test_params)
 
 
-def _resolve_param_path(test_path, rel_path):
+def _override_global_loaders(loaders):
+    if loaders is None:
+        return get_loaders()
+    else:
+        return {**get_loaders(), **loaders}
+
+def _resolve_param_path(test_path, rel_path, loaders):
     if rel_path:
         if is_iterable(rel_path):
             return [test_path.parent / p for p in rel_path]
@@ -357,7 +383,7 @@ def _resolve_param_path(test_path, rel_path):
 
     param_path_candidates = [
             test_path.with_suffix(x)
-            for x in get_loaders()
+            for x in loaders
     ]
     param_paths = [
             x for x in param_path_candidates if x.exists()
@@ -389,20 +415,24 @@ def _resolve_param_path(test_path, rel_path):
 
     return param_paths[0]
 
-@lru_cache()
-def _load_and_cache_suite_params(param_path):
-    if not param_path.exists():
-        err = ConfigError(
-                param_path=param_path,
-        )
-        err.brief = "can't find parametrization file"
-        err.blame += "'{param_path}' does not exist."
-        raise err
-
-    loaders = get_loaders()
+def _load_test_params(loaders, param_path, test_name):
+    loader = _pick_loader_by_suffix(loaders, param_path)
+    suite_params = _load_and_cache_suite_params(loader, param_path)
 
     try:
-        loader = loaders[param_path.suffix]
+        return suite_params[test_name]
+
+    except KeyError:
+        err = ConfigError(
+                test_name=test_name,
+        )
+        err.brief += "must specify parameters for '{test_name}'"
+        err.hints += "make sure the top-level data structure in the parameter file is a dictionary where the keys are the names of test functions, and the values are dictionaries of test parameters."
+        raise err from None
+
+def _pick_loader_by_suffix(loaders, param_path):
+    try:
+        return loaders[param_path.suffix]
 
     except KeyError:
         err = ConfigError(
@@ -417,6 +447,16 @@ def _load_and_cache_suite_params(param_path):
         err.blame += "the given extension is not recognized: {param_path.suffix}"
         raise err from None
 
+@lru_cache()
+def _load_and_cache_suite_params(loader, param_path):
+    if not param_path.exists():
+        err = ConfigError(
+                param_path=param_path,
+        )
+        err.brief = "can't find parametrization file"
+        err.blame += "'{param_path}' does not exist."
+        raise err
+
     try:
         return loader(param_path)
 
@@ -430,23 +470,13 @@ def _load_and_cache_suite_params(param_path):
         err2.blame += "{err}"
         raise err2 from None
 
-def _load_test_params(param_path, test_name):
-    suite_params = _load_and_cache_suite_params(param_path)
-
-    try:
-        return suite_params[test_name]
-
-    except KeyError:
-        err = ConfigError(
-                test_name=test_name,
-        )
-        err.brief += "must specify parameters for '{test_name}'"
-        err.hints += "make sure the top-level data structure in the parameter file is a dictionary where the keys are the names of test functions, and the values are dictionaries of test parameters."
-        raise err from None
-
-def _process_test_params(test_params_in, preprocess, schema):
+def _process_test_params(test_params_in, preprocess, context, schema):
     if preprocess:
-        test_params_in = preprocess(test_params_in)
+        sig = inspect.signature(preprocess)
+        if len(sig.parameters) > 1:
+            test_params_in = preprocess(test_params_in, context)
+        else:
+            test_params_in = preprocess(test_params_in)
 
     if not isinstance(test_params_in, list):
         raise ConfigError(
@@ -584,3 +614,8 @@ def _format_case_params(case_params):
     except:
         return repr(case_params)
 
+class Context:
+
+    def __init__(self, path, key):
+        self.path = path
+        self.key = key
